/*
 *  diskdumpfmt.c
 *
 *  Copyright (C) 2004-2006  FUJITSU LIMITED
 *  Written by Akira Imamura (imamura.akira@jp.fujitsu.com)
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: diskdump_format.c,v 1.71 2006/04/21 19:38:38 akira Exp $ */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <linux/swap.h>
#include <assert.h>
#include <time.h>

#include "diskdump_format.h"
#include "version.h"
#include "diskdumpcmn.h"
#include "diskdumpelf.h"
#include "dumpheader.h"

int symbol;	/* for progress bar */

static inline void set_err_msg(dump_device_t *device, const char *msg)
{
	asprintf(&device->err_msg, "%s failed: %s.", msg, strerror(errno));
}

/*
 * It returns the number of parameter values gotten from "/proc/diskdump".
 */
int get_module_params(dump_device_t *device)
{
	u64 val;
	const u64 invalid = (u64)-1;

	if ((val = scan_param_proc(SAMPLE_RATE)) == invalid) {
		return 0;		/* Got no parameter value. */
	}
	device->req->sample_rate = (int32_t)val;

	if ((val = scan_param_proc(BLOCK_ORDER)) == invalid) {
		return 1;		/* Got 1 parameter value. */
	}
	device->req->block_order = (int32_t)val;

	if ((val = scan_param_proc(TOTAL_BLOCKS)) == invalid) {
		return 2;		/* Got 2 parameter values. */
	}
	device->req->format_size = (off_t)val;

	if ((val = scan_param_proc(DUMP_LEVEL)) == invalid) {
		val = 0;		/* Kernel does not support
					   Partial dump. Use dump_level 0 */
	}
	device->req->dump_level = (int32_t)val;

	if ((val = scan_param_proc(COMPRESS)) == invalid) {
		val = 0;		/* Kernel does not support
					   Dump compression. Use compress 0 */
	}
	device->req->compress = (int32_t)val;

	return 3;			/* Got 3 parameter values. */
}

static int read_device(dump_device_t *device, off_t blk_nr)
{
	const off_t offset = blk_nr * device->page_size;

	if (device->block_size <= blk_nr) {
		asprintf(&device->err_msg, "block size should be more than %Lu",
			(u64)blk_nr);
		return FAILED;
	}

	if (lseek(device->fd, offset, SEEK_SET) == (off_t)-1) {
		set_err_msg(device, "lseek in read_device()");
		return FAILED;
	}

	if (read(device->fd, device->scratch, device->page_size)
	  < device->page_size) {
		set_err_msg(device, "read in read_device()");
		return FAILED;
	}

	return SUCCEEDED;
}

static int write_device(dump_device_t *device, off_t blk_nr)
{
	const off_t offset = blk_nr * device->page_size;

	if (device->block_size <= blk_nr) {
		asprintf(&device->err_msg, "block size should be more than %Lu",
			(u64)blk_nr);
		return FAILED;
	}

	if (lseek(device->fd, offset, SEEK_SET) == (off_t)-1) {
		set_err_msg(device, "lseek in write_device()");
		return FAILED;
	}

	if (write(device->fd, device->scratch, device->page_size) == -1) {
		set_err_msg(device, "write in write_device()");
		return FAILED;
	}

	return SUCCEEDED;
}

static off_t get_block_size(dump_device_t *device)
{
	struct stat stt;
	off_t block_size;
	const off_t failed = (off_t)-1;

	assert(device->page_size);

	if (fstat(device->fd, &stt) == -1) {
		set_err_msg(device, "fstat in get_block_size()");
		return failed;
	}

	if (S_ISBLK(stt.st_mode) || S_ISCHR(stt.st_mode)) {
		unsigned long sz;

		if (ioctl(device->fd, BLKGETSIZE, &sz) == -1) {
			set_err_msg(device, "ioctl in get_block_size()");
			return failed;
		}
		block_size = sz / (device->page_size / 512);
	} else {
		block_size = stt.st_size / device->page_size;
	}

	return block_size;	/* Got an exact block size. */
}

/*
 * It returns the value of diskdump format version (0 to n).
 */
static int get_header_params(dump_device_t *device, off_t blk_nr)
{
	if (read_device(device, blk_nr) == FAILED) {
		device->reason = "called by get_header_params()";
		return FAILED;
	}
	memcpy(device->dev, device->scratch, sizeof(format_params_t));

	if (strlen(device->dev->cmd_version) == 0) {
		/* command version not found in the parameter header. */
		memcpy(device->dev->cmd_version, "?", 2);
	}

	if (memcmp(device->dev->signature, DUMP_PARTITION_SIGNATURE, SIG_LEN)) {
		device->broken_header = true;
	}

	return device->dev->fmt_version;	/* Got exact parameters */
}

static inline bool is_full_formatted(dump_device_t *device)
{
	return  (device->dev->sample_rate == 0) &&
		(device->dev->block_order == 0) &&
		(device->dev->format_size == device->block_size);
}

static inline bool module_params_unknown(dump_device_t *device)
{
	return  (device->req->sample_rate == 0) &&
		(device->req->block_order == 0) &&
		(device->req->format_size == device->block_size);
}

static inline bool module_params_invalid(dump_device_t *device)
{
	return  (device->req->sample_rate < 0) &&
		(device->req->block_order < 0) &&
		(device->req->format_size < 0);
}

bool enough_block_size(dump_device_t *device)
{
	if (device->req->dump_level || device->req->compress) {
		/*
		* The device may not actually have enough block size.
		* The exact size of saved memory cannot be determined when
		* partial or compressed dump is activated.
		*/
		return true;
	} else {
		return (device->block_size >= device->req->format_size);
	}
}

static bool cmd_version_match(format_params_t *dev, format_params_t *req)
{
	struct version_form ver_dev, ver_req;

	if (convert_version_number(&ver_dev, dev->cmd_version) != 3) {
		return false;	/* Command version does not match. */
	}

	if (convert_version_number(&ver_req, req->cmd_version) != 3) {
		return false;	/* Command version does not match. */
	}

	return  (ver_dev.major == ver_req.major) &&
		(ver_dev.minor == ver_req.minor) &&
		(ver_dev.micro == ver_req.micro);
}

static inline bool fmt_params_satisfied(dump_device_t *device)
{
	return  (device->dev->sample_rate <= device->req->sample_rate) &&
		(device->dev->block_order == device->req->block_order) &&
		(device->dev->format_size == device->req->format_size) &&
		(cmd_version_match(device->dev, device->req));
}

static inline int check_swap_device(dump_device_t *device)
{
	if (device->has_panic_dump) {
		if (device->in_progress) {
			return SWAP_WITH_IN_PROGRESS_DUMP;
		} else {
			return SWAP_WITH_DUMP;
		}
	} else {
		return SWAP_FORMATTED;
	}
}

int check_device(dump_device_t *device, bool ignore)
{
	if (!enough_block_size(device)) {
		asprintf(&device->reason, "block size should be more than %Lu",
			(u64)device->req->format_size);
		return NOT_SUITABLE_DEVICE;
	}

	if (device->is_swap_device) {
		device->reason = "-";
		return check_swap_device(device);
		/*
		 * return one of the following
		 *   SWAP_FORMATTED
		 *   SWAP_WITH_DUMP
		 *   SWAP_WITH_IN_PROGRESS_DUMP
		 */
	}

	if (!ignore && module_params_unknown(device)) {
		device->reason = "module parameters unknown";
		return FORMAT_REQUIRED;
	}

	if (module_params_invalid(device)) {
		device->reason = "module parameters invalid";
		return NOT_SUITABLE_DEVICE;
	}

	if (!device->is_dump_device) {
		device->reason = "format with -f";
		return NOT_A_DUMP_DEVICE;
	}

	if (device->in_progress) {
		device->reason = "format as the need arise";
		return HAS_IN_PROGRESS_PANIC_DUMP;
	} else if (device->has_panic_dump) {
		device->reason = "format as the need arise";
		return HAS_PANIC_DUMP;
	}

	if (is_full_formatted(device)) {
		device->reason = "full formatted";
		return FORMATTED;
	}

	if (device->ver0_formatted) {
		device->reason = "v0-formatted";
		return FORMATTED;
	}

	if (fmt_params_satisfied(device)) {
		device->reason = "parameters satisfied";
		return FORMATTED;
	}

	if (device->broken_header) {
		device->reason = "broken parameter header";
		return FORMAT_REQUIRED;
	}

	/* Comparing the current parameters with the required ones */
	if (device->dev->fmt_version > CURRENT_DISKDUMP_FORMAT_VER) {
		asprintf(&device->reason, "unknown format version %d",
			device->dev->fmt_version);
	} else if (!cmd_version_match(device->dev, device->req)) {
		device->reason = "format version does not match";
	} else if (device->dev->sample_rate > device->req->sample_rate) {
		asprintf(&device->reason, "sample rate should be %d or less",
			device->req->sample_rate);
	} else if (device->dev->block_order != device->req->block_order) {
		asprintf(&device->reason, "block order should be %d",
			device->req->block_order);
	} else if (device->dev->format_size != device->req->format_size) {
		asprintf(&device->reason, "format size should be %Lu",
			(u64)device->req->format_size);
	} else {
		device->reason = "reason unknown";
	}

	return FORMAT_REQUIRED;
}

struct params_str {
	char	*fmt_version;
	char	*sample_rate;
	char	*block_order;
	char	*format_size;
};

static void determine_format_params(struct params_str *p, dump_device_t *device)
{
	if (device->dev->fmt_version > DISKDUMP_FORMAT_VER_0) {
		/* The previous parameters are found by a parameter header. */
		asprintf(&p->fmt_version, "%d-%s",
			device->dev->fmt_version, device->dev->cmd_version);
		asprintf(&p->sample_rate, "%d", device->dev->sample_rate);
		asprintf(&p->block_order, "%d", device->dev->block_order);
		asprintf(&p->format_size, "%Lu", (u64)device->dev->format_size);
	} else {
		/* Use "?" or "-" is used as the previous parameters
		 * if they are not found. */
		p->fmt_version = (device->ver0_formatted)? "0" : "?";
		p->sample_rate = (device->ver0_formatted)? "-" : "?";
		p->block_order = (device->ver0_formatted)? "-" : "?";
		p->format_size = (device->ver0_formatted)? "-" : "?";
	}
}

static bool device_has_panic_dump(dump_device_t *device)
{
	if (read_device(device, 1) == FAILED) {
		return false;
	}
	if (memcmp(device->scratch, DISK_DUMP_SIGNATURE,
				sizeof(DISK_DUMP_SIGNATURE) - 1) == 0) {
		return true;	/* device has panic dump */
	} else {
		return false;
	}
}

struct panic_dump_params {
	char	*node_name;
	char	*release;
	char	*version;
	char	*machine;
	char	*dumped_date;
	char	*passed_days;
	char	*dump_size;
};

static void determine_panic_dump_params(struct panic_dump_params *p, dump_device_t *device)
{
	struct disk_dump_header *header;

	header = malloc(sizeof(*header));

	if (device_has_panic_dump(device) && header) {
		/* The panic dump parameters are found by a disk dump header. */
		time_t	timeofdeath;

		memcpy(header, device->scratch, sizeof(*header));
		timeofdeath	= header->timestamp.tv_sec;
		p->node_name	= header->utsname.nodename;
		p->release	= header->utsname.release;
		p->machine	= header->utsname.machine;
		p->dumped_date	= get_date_time_str(timeofdeath);
		asprintf(&p->passed_days, "%d",
				diff_days(time(NULL), timeofdeath));
		asprintf(&p->dump_size, "%Lu",
			(u64)header->block_size * header->written_blocks);
	} else {
		/* "?" is used as the panic dump parameters
		 * if they are not found. */
		p->node_name	= "?";
		p->release	= "?";
		p->machine	= "?";
		p->dumped_date	= "?";
		p->passed_days	= "?";
		p->dump_size	= "?";
	}
}

static char *get_device_status(int status)
{
	switch (status) {
	case FORMATTED:
		return "formatted";
	case HAS_PANIC_DUMP:
		return "has panic dump";
	case NOT_A_DUMP_DEVICE:
		return "initial format required";
	case FORMAT_REQUIRED:
		return "format required";
	case NOT_SUITABLE_DEVICE:
		return "not suitable device";
	case SWAP_FORMATTED:
		return "swap device";
	case SWAP_WITH_DUMP:
		return "swap device with panic dump";
	case HAS_IN_PROGRESS_PANIC_DUMP:
		return "has incomplete panic dump";
	case SWAP_WITH_IN_PROGRESS_DUMP:
		return "swap device with incomplete dump";
	default:
		return "unknown";
	}
}

/*
 * print description of the status of the device.
 * status should be the return value of check_format().
 */
void show_device_info(dump_device_t *device, int status, bool verbose)
{
	struct params_str current;	/* Current Parameters */
	struct panic_dump_params pdump;

	if (!verbose) {
		return;	/* It is not necessary to do anymore. */
	}

	Info("%s\n", device->name);
	Info("  status     : %s\n", get_device_status(status));
	Info("  block size : %Lu\n", (u64)device->block_size);
	switch  (status) {
	case FORMATTED:
	case FORMAT_REQUIRED:
		determine_format_params(&current, device);
		Info("  version    : %s\n", current.fmt_version);
		Info("  sample rate: %s\n", current.sample_rate);
		Info("  block order: %s\n", current.block_order);
		Info("  format size: %s\n", current.format_size);
		break;
	case HAS_PANIC_DUMP:
	case HAS_IN_PROGRESS_PANIC_DUMP:
	case SWAP_WITH_DUMP:
	case SWAP_WITH_IN_PROGRESS_DUMP:
		determine_panic_dump_params(&pdump, device);
		Info("  node name  : %s\n", pdump.node_name);
		Info("  release    : %s\n", pdump.release);
		Info("  machine    : %s\n", pdump.machine);
		Info("  dumped date: %s\n", pdump.dumped_date);
		Info("  current    : %s\n", get_date_time_str(time(NULL)));
		Info("  passed days: %s\n", pdump.passed_days);
		Info("  dump size  : %s\n", pdump.dump_size);
		break;
	}
	Info("  note       : %s\n", device->reason);
}

void show_format_info(dump_device_t *device, bool verbose)
{
	struct params_str pre;	/* Previous Parameters */
	format_params_t post = {
		.fmt_version = device->req->fmt_version,
		.sample_rate = device->req->sample_rate,
		.block_order = device->req->block_order,
		.format_size = device->req->format_size
	};
	char *post_fmt_version;

	if (!verbose) {
		return;	/* It is not necessary to do anymore. */
	}

	if (!enough_block_size(device)) {
		post.format_size = device->block_size;
	}

	determine_format_params(&pre, device);

	if (module_params_unknown(device)) {
		asprintf(&post_fmt_version, "%d", DISKDUMP_FORMAT_VER_0);
	} else {
		asprintf(&post_fmt_version, "%d-%s",
			CURRENT_DISKDUMP_FORMAT_VER, CMD_VERSION);
	}

	Info("%s\n", device->name);
	Info("  block size : %Lu\n", (u64)device->block_size);
	Info("  version    : %s => %s\n", pre.fmt_version, post_fmt_version);
	Info("  sample rate: %s => %d\n", pre.sample_rate, post.sample_rate);
	Info("  block order: %s => %d\n", pre.block_order, post.block_order);
	Info("  format size: %s => %Lu\n",
					pre.format_size, (u64)post.format_size);
}

static int dump_is_in_progress(dump_device_t *device)
{
	struct disk_dump_header *header;

	if (read_device(device, 1) == FAILED) {
		device->reason = "read_device called in dump_is_in_progress()";
		return FAILED;
	}

	if ((header = malloc(sizeof(*header))) == NULL) {
		device->reason = "malloc called in dump_is_in_progress()";
		return FAILED;
	}

	memcpy(header, device->scratch, sizeof(*header));
	switch(header->status & ~DUMP_HEADER_COMPRESSED) {
	case DUMP_HEADER_INCOMPLETED:
	case DUMP_HEADER_IN_PROGRESS:
		return true;	/* panic dump is in progress */
	}

	return false;	/* panic dump is full */
}

static bool device_is_swap(dump_device_t *device)
{
	union swap_header *swh;

	if (read_device(device, 0) == FAILED) {
		device->reason = "read_device called in device_is_swap()";
		return false;		/* Reading the device failed. */
	}
	swh = (union swap_header *)device->scratch;

	if ((memcmp(swh->magic.magic,
		SWAP_PARTITION_SIGNATURE, SWAP_SIG_LEN) == 0)
			&& (swh->info.version == 1)) {
		return true;	/* swap partition */
	}

	return false;
}

/*
 * 1st word = 'disk'
 * 2nd word = 'dump'
 * 3rd word = block number [0...n]
 * 4th word = ((block number + 7) * 11)
 * 5th word = ((4th word+ 7)  * 11)
 * ..
 *
 * The device is checked out from the head of that.
 */
static bool check_block(dump_device_t *device, off_t blk_nr)
{
	int i;
	uint32_t *buf;
	uint32_t val = (uint32_t)blk_nr;

	if (read_device(device, blk_nr) == FAILED) {
		device->reason = "read_device called in check_block()";
		return false;		/* The block has NOT been formatted. */
	}
	buf = device->scratch;

	if (memcmp(buf, DUMP_PARTITION_SIGNATURE, SIG_LEN) != 0) {
		return false;		/* The block has NOT been formatted. */
	}

	for (i = 3; i < device->page_size / sizeof(int); i++) {
		val = Formula(val);
		if (buf[i] != val) {
			return false;	/* The block has NOT been formatted. */
		}
	}

	return true;			/* The block has been formatted. */
}

static inline bool device_is_dump_device(dump_device_t *device)
{
	return check_block(device, 0);
}

static inline bool device_has_param_header(dump_device_t *device)
{
	if (check_block(device, 2)) {
		/*
		 * The block #2 has been formatted. That means
		 * the parameter header is broken.
		 */
		return false;
	} else {
		return true;
	}
}

dump_device_t *alloc_device(void)
{
	dump_device_t	*device;

	if ((device = calloc(1, sizeof(*device))) == NULL) {
		return NULL;	/* Allocating the device failed. */
	}
	device->fd	= -1;
	device->reason	= "-";
	device->err_msg	= "[Unknown error]";

	if ((device->dev = calloc(1, sizeof(format_params_t))) == NULL) {
		return NULL;	/* Allocating the device failed. */
	}
	device->dev->fmt_version = -1;
	device->dev->sample_rate = -1;
	device->dev->block_order = -1;

	if ((device->req = calloc(1, sizeof(format_params_t))) == NULL) {
		return NULL;	/* Allocating the device failed. */
	}
	memcpy(device->req->signature, DUMP_PARTITION_SIGNATURE, SIG_LEN);
	device->req->fmt_version = CURRENT_DISKDUMP_FORMAT_VER;
	if (sizeof(CMD_VERSION) <= sizeof(device->req->cmd_version)) {
		memcpy(device->req->cmd_version, CMD_VERSION,
			sizeof(CMD_VERSION));
	} else {
		memcpy(device->req->cmd_version, CMD_VERSION,
			sizeof(device->req->cmd_version));
	}

	return device;		/* Allocating the device succeeded. */
}

int open_device(dump_device_t *device, const char *dev_name)
{
	extern int posix_memalign(void **, size_t, size_t);

	device->name = dev_name;

	if ((device->fd = open(dev_name, O_RDWR)) == -1) {
		set_err_msg(device, "open in open_device()");
		return FAILED;
	}

	if ((device->page_size = sysconf(_SC_PAGE_SIZE)) == -1) {
		set_err_msg(device, "sysconf in open_device()");
		return FAILED;
	}

	if (posix_memalign(&device->scratch, device->page_size,
	    device->page_size) != 0) {
		device->err_msg = "posix_memalign in open_device() failed";
		return FAILED;
	}

	device->block_size = get_block_size(device);
	if (device->block_size == (off_t)-1) {
		/* set_err_msg is called by get_block_size() */
		return FAILED;
	} else if (device->block_size == 0) {
		asprintf(&device->err_msg, "no block in %s", device->name);
		return FAILED;
	}
	device->req->format_size = device->block_size;

	device->is_swap_device = false;
	device->is_dump_device = false;
	if (device_is_swap(device)) {
		device->is_swap_device = true;
	} else if (device_is_dump_device(device)) {
		device->is_dump_device = true;
	}

	if (!device->is_swap_device && !device->is_dump_device) {
		/* no later process required */
		return SUCCEEDED;
	}

	if (device_has_panic_dump(device)) {
		device->has_panic_dump = true;
		switch (dump_is_in_progress(device)) {
		case FAILED:
			/* set_err_msg is called by dump_is_in_progress() */
			return FAILED;
		case true:
			device->in_progress = true;
			break;
		case false:
			device->in_progress = false;
			break;
		}
	} else {
		device->has_panic_dump = false;
	}

	if (device->is_swap_device) {
		/* no later process required */
		return SUCCEEDED;
	}

	device->broken_header = !device_has_param_header(device);
	device->ver0_formatted
		=   device->is_dump_device
		&& !device->has_panic_dump
		&&  device->broken_header;

	if (device->ver0_formatted || device->has_panic_dump) {
		/* no later process required */
		return SUCCEEDED;
	}

	if ((get_header_params(device, 2)) < DISKDUMP_FORMAT_VER_0) {
		/* set_err_msg is called by get_header_params() */
		return FAILED;
	}

	return SUCCEEDED;
}

static int make_header_block(dump_device_t *device, off_t blk_nr)
{
	if (read_device(device, blk_nr) == FAILED) {
		device->reason = "read_device called in make_header_block()";
		return FAILED;
	}
	memset(device->scratch, 0, device->page_size);

	device->req->block_number = (int32_t)blk_nr;

	memcpy(device->scratch, device->req, sizeof(format_params_t));

	if (write_device(device, blk_nr) == FAILED) {
		device->reason = "write_device called in make_header_block()";
		return FAILED;
	}
	sync();

	return SUCCEEDED;
}

/*
 * 1st word = 'disk'
 * 2nd word = 'dump'
 * 3rd word = block number [0...n]
 * 4th word = ((block number + 7) * 11)
 * 5th word = ((4th word + 7) * 11)
 * ..
 *
 * The device is formatted according to the above rule.
 */
static int format_one_block(dump_device_t *device, off_t blk_nr)
{
	uint32_t *words;
	int i;
	int val;

	if (read_device(device, blk_nr) == FAILED) {
		device->reason = "called by format_one_block()";
		return FAILED;
	}
	words = device->scratch;

	memcpy(words, DUMP_PARTITION_SIGNATURE, SIG_LEN);

	words[2] = (uint32_t)blk_nr;

	val = Formula(blk_nr);
	for (i = 3; i < device->page_size / sizeof(words[0]); i++) {
		words[i] = val;
		val = Formula(val);
	}

	if (write_device(device, blk_nr) == FAILED) {
		device->reason = "called by format_one_block()";
		return FAILED;
	}

	return SUCCEEDED;
}

/*
 * It returns the number of the blocks formatted last.
 * If completed, then it returns blk_end (a value of the block end).
 */
static off_t format_blocks(dump_device_t *device, off_t blk_ini, off_t blk_end)
{
	off_t blk;

	for (blk = blk_ini; blk < blk_end; blk++) {
		if (format_one_block(device, blk) == FAILED) {
			break;
		}
	}
	return blk;
}

static off_t format_range(dump_device_t *device, off_t blk_ini, off_t blk_end)
{
	off_t blk_nr, fmtd_blk;
	const off_t rate = 1 <<
		(device->req->sample_rate + device->req->block_order);
	const off_t seq = 1 << device->req->block_order;

	if (device->do_progress) {
		progress_bar(device->name, symbol, 0, blk_end);
	}

	for (blk_nr = 0; blk_nr < blk_end; blk_nr += rate) {
		off_t ini = max(blk_ini, blk_nr);
		off_t end = min(blk_end, blk_nr + seq);

		if ((fmtd_blk = format_blocks(device, ini, end)) < end) {
			return fmtd_blk;	/* Formatting failed. */
		}

		if (device->do_progress) {
			progress_bar(device->name, symbol, blk_nr, blk_end);
		}
	}

	if (device->do_progress) {
		progress_bar_post(device->name, symbol, blk_nr, blk_end);
	}

	sync();

	return blk_end;	/* Formatting succeeded. */
}

/*
 * Not in order to make an incomplete device by failure of diskdumpfmt,
 * block #0 (the head of a device) is particularly formatted last.
 * It returns the number of the blocks formatted last.
 */
off_t format_device(dump_device_t *device)
{
	off_t fmtd_blk = 0;		/* the number of formatted blocks */
	const off_t failed = (off_t)-1;
	const off_t block_size = (off_t)device->block_size;
	const off_t format_size = (off_t)device->req->format_size;
	const bool do_full_format = module_params_unknown(device);

	if (block_size <= MINIMUM_FORMAT_SIZE) {
		asprintf(&device->reason, "%s cannot be used as a dump device.",
			device->name);
		return failed;
	}

	if (do_full_format) {
		/* full format */

		if ((fmtd_blk = format_range(device, 3, block_size))
		  < block_size) {
			return fmtd_blk;	/* Formatting failed. */
		}
		/* In this case, the device does not
		 * have a header block particularly. */
		if (format_one_block(device, 2) == FAILED) {
			return 2;		/* Formatting failed. */
		}
	} else {
		/* quick format */
		const off_t seq = 1 << device->req->block_order;
		off_t body_size;

		body_size = min(format_size, block_size);

		if (body_size < seq) {
			device->reason = "block size is small for block order.";
			return failed;
		}
		if ((fmtd_blk = format_range(device, 3, body_size))
							  < body_size) {
			return fmtd_blk;	/* Formatting failed. */
		}

		/* format footer blocks */
		if (format_size < block_size) {
			if ((fmtd_blk = format_blocks(device, format_size - seq,
					  format_size)) < format_size) {
				return fmtd_blk;/* Formatting failed. */
			}
		}

		/* format the last blocks of the device */
		if ((fmtd_blk = format_blocks(device, block_size - seq,
				  block_size)) < block_size) {
			return fmtd_blk;	/* Formatting failed. */
		}

		if (make_header_block(device, 2) == FAILED) {
			return 2;		/* Formatting failed. */
		}
	}

	if (format_one_block(device, 1) == FAILED) {
		return 1;			/* Formatting failed. */
	}
	if (format_one_block(device, 0) == FAILED) {
		return 0;			/* Formatting failed. */
	}

	sync();

	return format_size;	/* Formatting succeeded. */
}
