/*
 *  diskdumpcmn.c
 *
 *  Copyright (C) 2005, 2006  FUJITSU LIMITED
 *
 */
/* $Id: diskdumpcmn.c,v 1.24 2006/04/21 19:38:38 akira Exp $ */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <termio.h>

#include "version.h"
#include "diskdumpcmn.h"

#define Stderr(x, ...)	fprintf(stderr, x, ## __VA_ARGS__)

static const u64 failed = (u64)-1;

#define DEFAULT_WINSZ	80

struct option {
	char	*kind;	/* character used as an option */
	char	*cmnt;	/* description to an option */
};

static void set_opt(struct option *opt, char *kind, char *comment)
{
	opt->kind = kind;
	opt->cmnt = comment;
}

static void print_opt(struct option *opt, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (opt[i].kind != NULL) {
			Stderr("\t%s:   %s\n", opt[i].kind, opt[i].cmnt);
		}
	}
}

void usage(const char *cmd)
{
	int n = 0;
	const char *opts;
	struct option opt[5] = {
		{NULL, NULL},
		{NULL, NULL},
		{NULL, NULL},
		{NULL, NULL},
		{NULL, NULL}
	};
	int ret;

	/*
	 * -v: some verbose outputs.
	 * -V: version information.
	 * Both of them are common options for each command.
	 */
	if (strcmp(cmd, "diskdumpfmt") == 0) {
		opts = "[-v] [[-a] [-p] [-f]]|[[-c] [-i]]";
		set_opt(&opt[n++], "-a",
		"format all blocks of the dump partition");
		set_opt(&opt[n++], "-c",
		"check the status of the dump partition");
		set_opt(&opt[n++], "-f",
		"format the dump partition forcibly");
		set_opt(&opt[n++], "-i",
		"ignore that the diskdump module is not loaded, work with -c");
		set_opt(&opt[n++], "-p",
		"display progress for formatting");
		ret = 255;
	} else if (strcmp(cmd, "savecore") == 0) {
		opts = "[-D] [-f] [-p] [-P pageflags] [-v] [-X]";
		set_opt(&opt[n++], "-D",
		"delete dump from the dump partition after saved");
		set_opt(&opt[n++], "-f",
		"create a dump file forcibly even if dump directory exists");
		set_opt(&opt[n++], "-p",
		"display progress for creating a dump file");
		set_opt(&opt[n++], "-P",
		"uncompress pages which have page flags specified by option");
		set_opt(&opt[n++], "-X",
		"delete dump from the dump partition without saving panic dump");
		ret = 1;
	} else if (strcmp(cmd, "diskdumpctl") == 0) {
		opts = "[-u] [-v]";
		set_opt(&opt[n++], "-u", "unregister the device");
		ret = 1;
	} else {
		Stderr("Error: command '%s' does not match.\n", cmd);
		exit(255);
	}

	Stderr("Usage: %s %s device\n", cmd, opts);
	Stderr("       %s -V\n", cmd);

	Stderr("Available options:\n");
	print_opt(opt, n);	/* shown as "\toption:   description" */
	Stderr("\t-v:   provide some verbose outputs\n");
	Stderr("\t-V:   show %s version and exit\n", cmd);

	exit(ret);
}

void version(const char *cmd)
{
	Stderr("%s version %s\n", cmd, CMD_VERSION);

	exit(0);
}

int convert_version_number(struct version_form *ver, const char *v_str)
{
	/* Initialize the version number as "0.0.0". */
	ver->major = 0;
	ver->minor = 0;
	ver->micro = 0;

	return sscanf(v_str, "%d.%d.%d", &ver->major, &ver->minor, &ver->micro);
}

/*
 * It gets some format parameters, sample_rate, block_order,
 * total_blocks and so on, from "/proc/diskdump".
 */
u64 scan_param_proc(const char *param_name)
{
	FILE *fp;
	char buf[BUFSIZ];
	u64 val;

	if ((fp = fopen(_PROC_DISKDUMP, "r")) == NULL) {
		return failed;
	}

	while (fgets(buf, BUFSIZ, fp) != NULL) {
		if (strstr(buf, param_name) != NULL) {
			if (sscanf(buf, "# %*s %Lu", &val) != 1) {
				fclose(fp);
				return failed;
			}
			fclose(fp);
			return val;	/* Got an exact value. */
		}
	}

	fclose(fp);
	return failed;
}

static inline int get_window_width(void)
{
	int rc;
	struct winsize winsz;

	rc = ioctl(0, TIOCGWINSZ, &winsz);
	if (rc < 0 || winsz.ws_col == 0) {
		return DEFAULT_WINSZ;
	} else {
		return winsz.ws_col;
	}
}

static inline void print_bar(int symbol, int len, int max)
{
	int i;

	if (!(0x21 <= symbol && symbol < 0x7f)) {
		return;
	}

	for (i = 0; i < len; i++) {
		Stderr("%c", symbol);
	}

	Stderr("%*c", max - len + 1, ' ');
}

static inline void warning_progress_bar(const char *act)
{
	static int dot = 0;
	int max = 3;

	Stderr("Warning: unable to %s progress bar correctly due to some error.", act);
	print_bar('.', dot, max);
	Stderr("\r");

	if (dot++ > max) {
		dot = 0;
	}
}

static void print_progress(const char *item, int symbol, off_t now, off_t max)
{
	double progress;
	int bar_len, bar_avl;
	char *prefix = NULL, *suffix = NULL;

	progress = (double)now * 100.0
		 / (double)max;

	if (progress > 100.0) {
		progress = 100.0;
	}

	if (asprintf(&prefix, "%s: ", item) < 0) {
		warning_progress_bar("print");
		return;
	}
	if (asprintf(&suffix, "[%5.1f%%]", progress) < 0) {
		warning_progress_bar("print");
		return;
	}

	bar_avl = get_window_width() - strlen(prefix) - strlen(suffix) - 1;
	bar_len = progress / 100 * bar_avl;

	Stderr(prefix);
	print_bar(symbol, bar_len, bar_avl);
	Stderr(suffix);
	Stderr("\r");
	free(prefix);
	free(suffix);
}

static inline void erase_bar(const char *item)
{
	int bar_avl;
	char *prefix = NULL, *suffix = NULL;

	if (asprintf(&prefix, "%s: ", item) < 0) {
		warning_progress_bar("erase");
		return;
	}
	if (asprintf(&suffix, "[%5.1f%%]", 100.0) < 0) {
		warning_progress_bar("erase");
		return;
	}

	bar_avl = get_window_width() - strlen(prefix);

	Stderr(prefix);
	Stderr("%*c", bar_avl, ' ');
	Stderr("\r");
	Stderr(prefix);
	Stderr(suffix);
	free(prefix);
	free(suffix);
}

void progress_bar(const char *item, int symbol, off_t now, off_t max)
{
	time_t tm;
	static time_t last_time = 0;

	tm = time(NULL);

	if (tm - last_time < 1) {
		return;
	}

	last_time = tm;

	print_progress(item, symbol, now, max);
}

void progress_bar_post(const char *item, int symbol, off_t now, off_t max)
{
	print_progress(item, symbol, now, max);
	erase_bar(item);
	Stderr("\n");
}

int execute_script(const char *script, const char *arg)
{
	struct stat sbuf;
	char *filename;
	char *cmdline;
	int res = -1;

	if ((filename = strdup(script)) == NULL) {
		return res;
	}

	if (stat(filename, &sbuf) == 0) {
		if (asprintf(&cmdline, "%s '%s'", filename, arg ? : "") < 0) {
			free(filename);
			return res;
		}
		res = system(cmdline);
		free(cmdline);
	}

	free(filename);
	return res;
}

int rm_dir(const char *path)
{
	char *rm;
	int ret;

	if (asprintf(&rm, "/bin/rm -rf %s", path) < 0) {
		return -1;
	}

	ret = system(rm);

	free(rm);
	return ret;
}

char *convert_ip_addr(uint32_t ip)
{
	char *ip_addr;

	if (asprintf(&ip_addr, "%d.%d.%d.%d",
			(ip >> 24) & 0xff,
			(ip >> 16) & 0xff,
			(ip >>  8) & 0xff,
			(ip      ) & 0xff) < 0) {
		return NULL;
	}

	return ip_addr;
}

bool is_swapon(void)
{
	FILE *fp;
	char line[BUFSIZ];

	if ((fp = fopen("/proc/swaps", "r")) == NULL)
		return false;

	if (fgets(line, BUFSIZ, fp) == NULL) {
		fclose(fp);
		return false;
	}

	/* to handle kernel-2.4.x */
	if (strncmp(line, "Filename", 8) == 0) { /* try again */
		if (fgets(line, BUFSIZ, fp) == NULL) {
			fclose(fp);
			return false;
		}
	}

	fclose(fp);
	return true;
}

int is_block_device(const char *device_name)
{
	struct stat stt;

	if (stat(device_name, &stt) < 0) {
		return -1;
	}

	return S_ISBLK(stt.st_mode);
}

static dev_t get_device_type(const char *device_name)
{
	struct stat stt;

	if (stat(device_name, &stt) < 0) {
		return (dev_t)-1;
	}

	return stt.st_rdev;
}

static int is_used(dev_t device_type, const char *filename)
{
	FILE *fp;
	char buf[BUFSIZ];
	char dev_name[BUFSIZ];
	dev_t dev_type = 0;
	enum {
		open_failed = -1,
		not_used,
		used
	};
	int ret = not_used;

	if ((fp = fopen(filename, "r")) == NULL) {
		return open_failed;
	}

	while (fgets(buf, BUFSIZ, fp) != NULL) {
		if (sscanf(buf, "%s %*s", dev_name) != 1) {
			continue;
		}
		if (!is_block_device(dev_name)) {
			continue;
		}
		if ((dev_type = get_device_type(dev_name)) == (dev_t)-1) {
			continue;
		}
		if (dev_type == device_type) {
			ret = used;
			break;
		}
	}

	fclose(fp);
	return ret;
}

int is_mounted(const char *device_name)
{
	dev_t device_type;
	int i;
	int ret;
	enum {
		ERROR = -1,
		NOT_MOUNTED,
		MOUNTED,
	};
	const char *file[] = {
		"/proc/mounts",	/* usually enough for kernel-2.6.x */
		"/etc/mtab"	/* able to get device that is on rootfs
				   on kernel-2.4.x */
	};

	if ((device_type = get_device_type(device_name)) == (dev_t)-1) {
		return ERROR;
	}

	for (i = 0; i < 2; i++) {
		ret = is_used(device_type, file[i]);
		if (ret == -1) {
			return ERROR;
		} else if (ret) {
			return MOUNTED;
		}
	}

	return NOT_MOUNTED;
}

char *get_date_time_str(const time_t t)
{
	struct tm *tm;
	char date[] = DATE_TIME_FORMAT, *date_time;

	tm = localtime(&t);
	strftime(date, sizeof(date), "%Y-%m-%d-%H:%M", tm);

	if ((date_time = strdup(date)) < 0) {
		return NULL;
	}

	return date_time;
}

int diff_days(time_t tm1, time_t tm2)
{
	return difftime(tm1, tm2) / (60 * 60 * 24);
}
