/*
 *  savecore.c
 *
 *  Copyright (C) 2004-2006  FUJITSU LIMITED
 *  Copyright (C) 2001-2005  Red Hat, Inc
 *
 *  Written by Nobuhiro Tachino (tachino@jp.fujitsu.com) based on
 *  netdump-server code.
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: savecore.c,v 1.70 2006/06/29 18:16:43 akira Exp $ */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <zlib.h>

#include "savecore.h"

#define CMD	"savecore"

#ifdef FORCE_SINGLE_PTLOAD
#define _FORCE_SINGLE_PTLOAD	true
#else
#define _FORCE_SINGLE_PTLOAD	false
#endif

static bool prog_flg = false;	/* becomes true if option '-p' specified. */
static bool frc_flg = false;	/* becomes true if option '-f' specified. */
static unsigned long long pagetype_flg; /* if matches, page is uncompressed */

static const int symbol = '#';	/* for progress bar */

static int nr_cpus;

bool generate_elf_header(DumpDevice *);


static DumpDevice *create_new_device(const char *device_name)
{
	DumpDevice *device;

	if ((device = calloc(1, sizeof(DumpDevice))) == NULL) {
		return NULL;
	}

	if ((device->name = strdup(device_name)) == NULL) {
		return NULL;
	}

	device->fd = -1;

	/* Lame default. should be read from kernel */
	device->page_offset = PAGE_OFFSET;

	return device;
}

static int open_device_file(DumpDevice *device)
{
	int fd;

	if ((fd = open(device->name, O_RDONLY)) < 0)
		return FAILED;

	device->fd = fd;
	return SUCCEEDED;
}

static int check_fs(DumpDevice *device)
{
	struct statfs buf;
	DumpFile *dump_file = device->dump_file;
	const struct disk_dump_header *header = device->header;
	off_t memdump_size;
	bool tried_once = false;
	enum {
		check_failed = -1,
		available,
		no_space
	};

redo:
	if (fstatfs(dump_file->fd, &buf) < 0)
		return check_failed;

	memdump_size = (off_t)header->written_blocks + 3; /* elf header */
	memdump_size *= header->block_size;
	memdump_size = (memdump_size + (buf.f_bsize - 1)) / buf.f_bsize;

	if (memdump_size <= buf.f_bavail)
		return available;

	if (!tried_once++)
		if (execute_script(DISKDUMP_NOSPACE, dump_file->dir) > 0)
			goto redo;

	return no_space;
}

static char *get_dir(DumpDevice *device)
{
	char *path;
	const char *ip_addr;
	char *date;

	if ((ip_addr = convert_ip_addr(DISKDUMP_IP)) == NULL) {
		return NULL;
	}
	if ((date = get_date_time_str(device->header->timestamp.tv_sec))
		== NULL) {
		return NULL;
	}
	if (asprintf(&path, "%s%s-%s/", DISKDUMP_DIR_PREFIX, ip_addr, date)
			< 0) {
		return NULL;
	}

	if (frc_flg) {
		if (rm_dir(path) < 0) {
			free(path);
			return NULL;
		}
	}

	if (mkdir(path, S_IRWXU) < 0) {
		free(path);
		return NULL;
	}

	return path;
}

static inline int get_bit(char *map, int byte, int bit)
{
	return map[byte] & (1<<bit);
}

static inline void set_bit(char *map, int byte, int bit)
{
	map[byte] |= 1<<bit;
}

static int next_bit(DumpDevice *device)
{
	char *src = device->bitmap + device->bitmap_len / 2;
	int val;

	val = get_bit(src, device->byte, device->bit);
	if (++device->bit == 8) {
		device->byte++;
		device->bit = 0;
	}

	return val;
}

static void expand_bitmap(DumpDevice *device)
{
	int byte, bit;

	for (byte = 0; byte < device->bitmap_len / 2; byte++)
		for (bit = 0; bit < 8; bit++)
			if (get_bit(device->bitmap, byte, bit)
			 && next_bit(device))
				set_bit(device->dumpable_bitmap, byte, bit);
}

static int read_device_header(DumpDevice *device)
{
	struct disk_dump_header *header;
	struct disk_dump_sub_header *sub_header;
	int ret;
	int bitmap_len;
	const int block_size = (int)sysconf(_SC_PAGESIZE);
	int sub_header_len;
	off_t offset;

	if (block_size < 0)
		return BLOCK_SIZE_INVALID;

	if ((header = malloc(block_size)) == NULL) {
		return UNABLE_TO_ALLOCATE;
	}

	offset = (off_t)block_size;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	if (read(device->fd, header, block_size) < block_size) {
		ret = READ_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	/* validate dump header */
	if (memcmp(header->signature, DISK_DUMP_SIGNATURE,
				sizeof(header->signature))) {
		ret = NO_PANIC_DUMP_HEADER;
		goto err;
	}

	if (header->block_size != block_size) {
		ret = SIZE_NOT_MATCH;
		goto err;
	}

	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size
	 || header->nr_cpus <= 0) {
		nr_cpus = header->nr_cpus;
		ret = NUMBER_OF_CPUS_INVALID;
		goto err;
	}

	/* read sub header */
	offset = (off_t)block_size * 2;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	sub_header_len = sub_header_length(header);
	if ((sub_header = malloc(sub_header_len)) == NULL) {
		ret = UNABLE_TO_ALLOCATE;
		goto err;
	}
	if (read(device->fd, sub_header, sub_header_len)
	  < sub_header_len) {
		ret = READ_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	device->sub_header = sub_header;
	device->sub_header_len = sub_header_len;

	/* read memory bitmap */
	bitmap_len = block_size * header->bitmap_blocks;
	device->bitmap_len = bitmap_len;

	offset = (off_t)block_size * (2 + header->sub_hdr_size);
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	if ((device->bitmap = malloc(bitmap_len)) == NULL) {
		ret = UNABLE_TO_ALLOCATE;
		goto err;
	}
	if ((device->dumpable_bitmap = calloc(1, bitmap_len)) == NULL) {
		ret = UNABLE_TO_ALLOCATE;
		goto err;
	}
	if (read(device->fd, device->bitmap, bitmap_len) < bitmap_len) {
		ret = READ_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	if (dump_is_partial(header))
		expand_bitmap(device);
	else
		memcpy(device->dumpable_bitmap, device->bitmap, bitmap_len);

	/*
	 * page 0: device header
	 * page 1: dump header
	 * page 2..: dump sub header
	 * page 2 + size of subheader: memory page 0
	 */
	device->device_page_start
		= (2 + header->sub_hdr_size + header->bitmap_blocks)
		* header->block_size;

	device->header = header;

	return COMPLETE;

err:
	free(header);
	return ret;
}

static bool pages_are_zero(int nr, int len, DumpDevice *device)
{
	int i;
	const unsigned int maxnr = nr + min(len, device->header->max_mapnr);

	for (i = nr >> 3; i < (maxnr + 7) >> 3; i++)
		if (device->bitmap[i])
			return false;
	return true;
}

static void build_segments_info(DumpDevice *device)
{
	const struct disk_dump_header *header = device->header;
	struct mem_segment	*seg, *seg_base;
	unsigned int chunk;			/* = HOLE_SIZE blocks */
	const unsigned int max_mapnr = header->max_mapnr;
	const unsigned int max_segments = max_mapnr / HOLE_SIZE / 2;
	bool in_segment;

	if ((seg = seg_base = calloc(max_segments, sizeof(struct mem_segment)))
			== NULL) {
		return;
	}

	if (_FORCE_SINGLE_PTLOAD || header->max_mapnr <= HOLE_SIZE) {
		seg->blk = 0;
		seg->len = device->header->max_mapnr;

		device->num_segments = 1;
		device->segments = seg_base;

		return;
	}

	/* The first segment must have the first block(blk#0). */
	seg = seg_base;
	seg->blk = 0;
	in_segment = true;
	for (chunk = 1; chunk < divideup(max_mapnr, HOLE_SIZE); chunk++) {
		if (pages_are_zero(chunk * HOLE_SIZE, HOLE_SIZE, device)) {
			if (in_segment) {
				seg->len = chunk * HOLE_SIZE - seg->blk;
				in_segment = false;
			}
		} else {
			if (!in_segment) {
				seg++;
				seg->blk = chunk * HOLE_SIZE;
				in_segment = true;
			}
		}
	}
	/*
	 * The last segment must contain the last blocks even if
	 * they are zero-filled pages.
	 */
	seg->len = max_mapnr - seg->blk;

	device->num_segments = seg + 1 - seg_base;
	device->segments = seg_base;
}

static bool page_is_in_segs(unsigned int nr, DumpDevice *device)
{
	struct mem_segment *seg;
	int i;

	for (i = 0; i < device->num_segments; i++) {
		seg = &device->segments[i];

		if (seg->blk <= nr && nr < seg->blk + seg->len)
			return true;
	}
	return false;
}

/******************************************************************************
 *                       compressed dump                                      *
 ******************************************************************************/
static inline bool dump_is_compressed(DumpDevice *device)
{
	return device->header->status & DUMP_HEADER_COMPRESSED;
}

static inline bool page_is_dumpable(unsigned int nr, DumpDevice *device)
{
	return device->dumpable_bitmap[nr>>3] & (1 << (nr & 7));
}

static inline bool write_compress_header(DumpDevice *device)
{
	DumpFile *dump_file = device->dump_file;
	int size;
	const struct disk_dump_header *header = device->header;
	int i;
	unsigned long num_pages;

	/* write dump header */
	dump_file->data_offset = 0;
	if (write_blocks(dump_file, header, header->block_size))
		return FAILED;

	/* write dump sub header */
	if (write_blocks(dump_file, device->sub_header, device->sub_header_len))
		return FAILED;

	/* write bitmap */
	if (dump_is_partial(header)) {
		size = device->bitmap_len / 2;
		/* write 1st-bitmap */
		if (write_blocks(dump_file, device->bitmap, size))
			return FAILED;
		if (write_blocks(dump_file, device->dumpable_bitmap, size))
			return FAILED;
	} else {
		size = device->bitmap_len;
		if (write_blocks(dump_file, device->bitmap, device->bitmap_len))
			return FAILED;
	}

	dump_file->page_desc_offset = dump_file->data_offset;

	/* calculate data offset */
	num_pages = 0;
	for (i = 0; i <= header->max_mapnr; i++)
		if (page_is_dumpable(i, device))
			num_pages++;
	dump_file->data_offset += num_pages * sizeof(page_desc_t);

	return SUCCEEDED;
}

static int write_compress_pages(DumpDevice *device, unsigned char *scratch[])
{
	DumpFile *dump_file = device->dump_file;
	struct dump_page_header ph;
	int ret;
	int block_size = device->header->block_size;
	unsigned long len;
	unsigned char *rawdata = scratch[0], *data = scratch[1];

	/* read page header */
	if (read(device->fd, &ph, sizeof(ph)) != sizeof(ph))
		return READ_ERROR_IN_DUMP_DEVICE;

	/* sanity check */
	if (ph.size > block_size)
		return INVALID_SIZE_IN_PAGE_HEADER;

	/* try to read in the dump page itself */
	if (read(device->fd, rawdata, ph.size) < ph.size)
		/* if this fails, we may have run out of space */
		return READ_ERROR_IN_DUMP_DEVICE;

	/* write out the dump page header */
	if (dump_file->page_desc_offset > dump_file->data_offset)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (ph.flags & DUMP_DH_COMPRESSED && ph.page_flags & pagetype_flg) {
		ret = write_page_desc(dump_file,
			      &(page_desc_t){
				.offset     = dump_file->data_offset,
				.size       = block_size,
				.flags      = ph.flags & ~DUMP_DH_COMPRESSED,
				.page_flags = ph.page_flags
			      });
		if (ret)
			return ret;

		len = block_size;
		ret = uncompress(data, &len, rawdata, ph.size);
		if (ret != Z_OK || len != block_size)
			return UNCOMPRESS_ERROR;

		ret = write_page_data(dump_file, data, len);
	} else {
		ret = write_page_desc(dump_file,
			      &(page_desc_t){
				.offset     = dump_file->data_offset,
				.size       = ph.size,
				.flags      = ph.flags,
				.page_flags = ph.page_flags
			      });
		if (ret)
			return ret;

		/* try to write out the dump page itself */
		ret = write_page_data(dump_file, rawdata, ph.size);
	}
	if (ret)
		return ret;

	return 0;
}

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static inline int write_dump_header(DumpDevice *device)
{
	if (dump_is_compressed(device))
		return write_compress_header(device);

	if (generate_elf_header(device))
		return FAILED;
	else
		return SUCCEEDED;
}

static inline bool check_zero_page(DumpDevice *device, unsigned char *data)
{
	int i;

	for (i = 0; i < device->header->block_size; i++)
		if (data[i])
			return false;
	return true;
}

static inline bool page_is_ram(unsigned int nr, DumpDevice *device)
{
	return get_bit(device->bitmap, nr >> 3, nr & 7);
}

static inline bool dump_is_short_area(DumpDevice *device)
{
	return (device->header->status == DUMP_HEADER_SHORT_AREA);
}

static int write_pages(DumpDevice *device)
{
	DumpFile *dump_file = device->dump_file;
	int i;
	unsigned char *data[2] = {NULL, NULL};
	int ret = COMPLETE;
	const struct disk_dump_header *header = device->header;
	const int block_size = header->block_size;
	const char *device_name = device->name;
	off_t offset;
	const off_t failed = (off_t)-1;

	if ((data[0] = malloc(block_size)) == NULL) {
		ret = UNABLE_TO_ALLOCATE;
		goto out;
	}
	if ((data[1] = malloc(block_size)) == NULL) {
		ret = UNABLE_TO_ALLOCATE;
		goto out;
	}

	offset = (off_t)device->device_page_start;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_DEVICE;
		goto out;
	}

	offset = dump_file->data_offset;
	if (dump_seek(dump_file, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	for (i = 0; i < header->max_mapnr; i++) {
		if (prog_flg)
			progress_bar(device_name, symbol, i, header->max_mapnr);

		if (!page_is_in_segs(i, device))
			continue;

		if (dump_is_compressed(device)) {
			if (page_is_dumpable(i, device))
				if ((ret = write_compress_pages(device, data)))
					goto out;
		} else {
			if (page_is_dumpable(i, device)) {
				if (read(device->fd, data[0], block_size)
				  < block_size) {
					ret = READ_ERROR_IN_DUMP_DEVICE;
					goto out;
				}
				if (check_zero_page(device, data[0]))
					goto skip;

				if (write(dump_file->fd, data[0], block_size)
				  < block_size) {
					ret = WRITE_ERROR_IN_DUMP_FILE;
					goto out;
				}
			} else {
skip:
				if (lseek(dump_file->fd, block_size, SEEK_CUR) == failed) {
					ret = SEEK_ERROR_IN_DUMP_FILE;
					goto out;
				}
			}
		}
	}

	/*
	 * Set the crash dump file size intentionally because the last block
	 * may not be actually written.
	 */
	if ((offset = lseek(dump_file->fd, 0, SEEK_CUR)) == failed) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	if (ftruncate(dump_file->fd, offset) < 0) {
		ret = DUMP_FILE_SIZE_UNDETERMINED;
		goto out;
	}

	if (prog_flg)
		progress_bar_post(device_name, symbol, i, header->max_mapnr);

out:
	free(data[0]);
	free(data[1]);
	if (dump_is_compressed(device))
		device->header->status &= ~DUMP_HEADER_COMPRESSED;
	return ret;
}

static int blank_out_signature(DumpDevice *device)
{
	int fd;
	off_t offset;
	const struct disk_dump_header *header = device->header;
	const int block_size = header->block_size;

	/* reopen device file */
	if (close(device->fd) < 0)
		return FAILED;
	if ((fd = open(device->name, O_RDWR)) < 0)
		return FAILED;
	device->fd = fd;

	offset = (off_t)block_size;
	if (device_seek(device, offset) < 0)
		return FAILED;

	if (write(device->fd, ALREADY_SAVED_SIGNATURE, SIG_LEN) < SIG_LEN)
		return FAILED;

	return SUCCEEDED;
}

static inline void set_pagetype_flag(char *val)
{
	pagetype_flg = strtoull(val, NULL, 0);
}

static inline bool memory_page_completed(DumpDevice *device)
{
	return (device->header->status == DUMP_HEADER_COMPLETED);
}

int main(int argc, char *argv[])
{
	DumpDevice *device;
	DumpFile *dump_file = NULL;
	int opt;
	bool verbose = false, del_flg = false, delete_only = false;
	int ret;
	char *dir;

	while ((opt = getopt(argc, argv, "DfvVpP:X")) != -1) {
		switch (opt) {
		case 'D':
			del_flg = true;
			break;
		case 'f':
			frc_flg = true;
			break;
		case 'p':
			prog_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'P':
			set_pagetype_flag(optarg);
			break;
		case 'V':
			version(CMD);
			break;
		case 'X':
			del_flg = true;
			delete_only = true;
			break;
		default:
			usage(CMD);
			break;
		}
	}

	if (optind + 1 != argc)
		usage(CMD);

	if ((device = create_new_device(argv[optind])) == NULL) {
		Err("unable to create a new device.\n");
		exit(1);
	}

	if (open_device_file(device) < 0) {
		Syserr("open %s", device->name);
		exit(1);
	}
	Info("%s\n", device->name);

	ret = read_device_header(device);
	switch (ret) {
	case COMPLETE:
		break;
	case BLOCK_SIZE_INVALID:
		Err("block size in the dump header is invalid.\n");
		exit(1);
	case SEEK_ERROR_IN_DUMP_HEADER:
		Syserr("seek the dump header");
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Syserr("read the dump header");
		exit(1);
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n",
				device->name);
		exit(1);
	case SIZE_NOT_MATCH:
		Err("block size in the dump header does not match with system page size.\n");
		exit(1);
	case NUMBER_OF_CPUS_INVALID:
		Err("invalid nr_cpus value: %d\n", nr_cpus);
		exit(1);
	case SEEK_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("seek the dump sub header");
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("read the dump sub header");
		exit(1);
	case SEEK_ERROR_IN_MEMORY_BITMAP:
		Syserr("seek the memory bitmap");
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Syserr("read the memory bitmap");
		exit(1);
	case UNABLE_TO_ALLOCATE:
		Err("unable to allocate memory to read the device header.");
		exit(1);
	default:
		Err("unknown error (reading the device header)\n");
		exit(1);
	}
	Info("  device header : read\n");

	if (del_flg && delete_only) {
		goto del_sig;
	}

	dir = get_dir(device);
	if (!dir) {
		Err("unable to make the dump directory\n");
		exit(1);
	}

	dump_file = open_dump_file(dir);
	if (!dump_file) {
		Syserr("open the crash dump file");
		exit(1);
	}

	dump_file->device = device;
	device->dump_file = dump_file;
	Info("  dump file name: %s\n", dump_file->name);

	ret = check_fs(device);
	switch (ret) {
	case -1:
		Err("unable to check %s.\n", dump_file->name);
		exit(1);
	case 0:
		break;
	case 1:
		Err("no space for dump image.\n");
		exit(1);
	default:
		Err("unknown error (checking file system)\n");
		exit(1);
	}
	Info("  file system   : available\n");

	build_segments_info(device);
	Info("  segment count : %d\n", device->num_segments);

	if (write_dump_header(device) < 0) {
		Err("unable to write the dump header.\n");
		exit(1);
	}
	Info("  dump header   : written\n");

	if (!prog_flg)
		Info("  memory dump   : ");
	ret = write_pages(device);
	switch (ret) {
	case COMPLETE:
		break;
	case SEEK_ERROR_IN_DUMP_DEVICE:
		Syserr("seek the dump device");
		exit(1);
	case READ_ERROR_IN_DUMP_DEVICE:
		Syserr("read the dump device");
		exit(1);
	case WRITE_ERROR_IN_DUMP_FILE:
		Syserr("write to the crash dump file");
		exit(1);
	case SEEK_ERROR_IN_DUMP_FILE:
		Syserr("seek the crash dump file");
		exit(1);
	case DUMP_FILE_SIZE_UNDETERMINED:
		Syserr("set the crash dump file size");
		exit(1);
	case UNCOMPRESS_ERROR:
		Syserr("uncompress the crash dump file");
		exit(1);
	case UNABLE_TO_ALLOCATE:
		Err("unable to allocate memory to write to pages.");
		exit(1);
	default:
		Err("unknown error (writing memory dump)\n");
		exit(1);
	}
	if (!prog_flg)
		Info("written\n");
	else
		Info("  memory dump   : written\n");
	Info("  written blocks: %u\n", device->header->written_blocks);

del_sig:
	if (del_flg) {
		if (is_swapon()) {
			Err("swap is running.\n");
			exit(1);
		}
		if (blank_out_signature(device) == FAILED) {
			Syserr("blank out dump signature");
			exit(1);
		}
		Info("  dump signature: blanked out\n");
		if (delete_only) {
			exit(0);
		}
	}

	if (!memory_page_completed(device)) {
		Err("dump partition has incomplete memory image.\n");
		if (dump_is_short_area(device))
			exit(0);
		else
			exit(1);
	}

	if (rename_dump_file(dump_file) != 0) {
		Syserr("rename the crash dump file");
		exit(1);
	}
	Info("  dump file name: %s\n", dump_file->name);

	Info("  script to run : %s\n", DISKDUMP_SUCCESS);
	Info("  script status :\r");
	if (execute_script(DISKDUMP_SUCCESS, dump_file->dir) == 0) {
		Info("  script status : complete\n");
	} else {
		Info("  script status : incomplete or not found\n");
	}

	Info("saving panic dump completed.\n");
	exit(0);
}
